#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['phx_events']

package_data = \
{'': ['*']}

setup(name='phx-events',
      version='0.2.0',
      description='AsyncIO Phoenix Channels client for fast asynchronous message handling',
      author=None,
      author_email='Jethro Muller <git@jethromuller.co.za>',
      url=None,
      packages=packages,
      package_data=package_data,
      python_requires='>=3.9,<4',
     )
